import numpy as np
import pathlib

from matplotlib.axes import Axes
from PyQt5 import Qt
from typing import List, Optional, Union

from .window_base import MainWindow, SceneDictType
from .utils.types import contains_scene


__all__ = ['display_dataset']


class MainWindowDataset(MainWindow):

    def __init__(
        self,
        scenes: List[SceneDictType],
        fps: float,
        video_directory: Optional[Union[pathlib.Path, str]] = None,
        horizontal: bool = True,
        loop: bool = True,
        show_labels: bool = False,
        use_scene_cam: bool = False,
        debug: bool = False,
        parent: Qt.QWidget = None
    ):
        self.scenes = scenes
        self.fps = fps
        if len(scenes) == 0:
            return
        self.__frame_idx = 0
        scene_dict = self.scenes[0]

        image_keys = [key for key, value in scene_dict.items() if isinstance(value, np.ndarray)]
        scene_keys = [key for key, value in scene_dict.items() if contains_scene(value)]
        figure_keys = [key for key, value in scene_dict.items() if isinstance(value, Axes)]
        super(MainWindowDataset, self).__init__(
            image_keys=image_keys,
            scene_keys=scene_keys,
            figure_keys=figure_keys,
            video_directory=video_directory,
            horizontal=horizontal,
            show_labels=show_labels,
            use_scene_cam=use_scene_cam,
            debug=debug,
            parent=parent
        )
        self.render_(scene_dict, resetcam=True)

        # Setup looping timer and connect it to render a new scene at every timer callback.
        # If loop is true, start looping right away.
        self.timer = Qt.QTimer(self)
        self.timer.timeout.connect(self.render_loop)  # noqa
        if loop:
            self.toggle_looping()

    def render_loop(self):
        self.render_current_index()
        self.__frame_idx += 1  # TODO: run forward or backward

    def render_current_index(self, resetcam: bool = False) -> None:
        if 0 <= self.__frame_idx < len(self.scenes):
            scene_dict = self.scenes[self.__frame_idx]
            self.render_(scene_dict, resetcam=resetcam)
        else:
            # Reset frame index to a valid index.
            self.__frame_idx = max(min(self.__frame_idx, len(self.scenes) - 1), 0)
            # Stop the timer, if it is running.
            if self.timer.isActive():
                self.timer.stop()

    def _on_key(self, event_dict) -> None:
        super(MainWindowDataset, self)._on_key(event_dict)
        key_pressed = event_dict["keyPressed"]
        if key_pressed == "s":
            self.toggle_looping()
        elif key_pressed == "n" and not self.timer.isActive():
            self.__frame_idx += 1
            self.render_current_index()
        elif key_pressed == "b" and not self.timer.isActive():
            self.__frame_idx -= 1
            self.render_current_index()

    @staticmethod
    def print_usage():
        MainWindow.print_usage()
        print("\ts: play / pause",
              "\n\tn: next",
              "\n\tb: previous")

    def toggle_looping(self):
        if self.timer.isActive():
            self.timer.stop()
        else:
            self.timer.start(int(1 / self.fps * 1000))  # in milliseconds


def display_dataset(
    scenes: List[SceneDictType],
    horizontal: bool = True,
    loop: bool = False,
    fps: float = 30.0,
    video_directory: Optional[Union[pathlib.Path, str]] = None,
    show_labels: bool = False,
    use_scene_cam: bool = False,
    debug: bool = False
):
    """Display scenes stored in scene iterator as PyQt app.

    The scene iterator yields a dictionary that describes the elements of the scene, one dictionary per frame.
    See SceneDictType for currently supported object types.

    Args:
        scenes: sorted list of scene dictionaries.
        horizontal: window orientation, horizontal or vertical stacking.
        loop: start looping from beginning.
        fps: frames (i.e. scenes) per second for looping.
        video_directory: directory for storing screenshots.
        show_labels: display the scene dict keys.
        use_scene_cam: use camera transform from trimesh scenes.
        debug: printing debug information.
    """
    app = Qt.QApplication([])
    window = MainWindowDataset(
        scenes=scenes,
        fps=fps,
        video_directory=video_directory,
        horizontal=horizontal,
        loop=loop,
        show_labels=show_labels,
        debug=debug,
        use_scene_cam=use_scene_cam
    )
    app.aboutToQuit.connect(window.on_close)
    app.exec_()



import balanna.camera_trajectories as camera_trajectories
import balanna.utils as utils
import balanna.trimesh as trimesh

from balanna.rendering_dataset import render_dataset
from balanna.window_dataset import display_dataset
from balanna.window_generator import display_generated, display_scenes
from balanna.window_real_time import display_real_time, RealTimeNode
from balanna.window_base import SceneDictType


__all__ = [
  "camera_trajectories", 
  "trimesh",
  "display_dataset",
  "display_scenes",
  "display_generated",
  "display_real_time", 
  "RealTimeNode",
  "render_dataset",
  "SceneDictType",
]



import datetime
import image_grid
import importlib.metadata
import numpy as np
import packaging.version
import pathlib
import trimesh.path.entities
import trimesh.viewer
import time
import vedo

import matplotlib
from matplotlib.axes import Axes
from matplotlib.figure import Figure
from matplotlib.backends.backend_qtagg import (
    FigureCanvasQTAgg as FigureCanvas, NavigationToolbar2QT as NavigationToolbar)

from functools import partial
from PIL import Image
from PyQt5 import Qt
from typing import Any, Dict, List, Optional, Union
from vtk.qt.QVTKRenderWindowInteractor import QVTKRenderWindowInteractor

from balanna.utils.vedo_utils import trimesh_scene_2_vedo

SceneDictType = Dict[str, Union[trimesh.Scene, Axes, str, np.ndarray, vedo.Volume]]


class MainWindow(Qt.QMainWindow):

    def __init__(
        self,
        image_keys: List[str],
        scene_keys: List[str],
        figure_keys: List[str],
        video_directory: Optional[Union[pathlib.Path, str]] = None,
        horizontal: bool = True,
        show_labels: bool = False,
        use_scene_cam: bool = False,
        debug: bool = False,
        parent: Qt.QWidget = None
    ):
        Qt.QMainWindow.__init__(self, parent)
        frame = Qt.QFrame()
        if horizontal:
            vl = Qt.QHBoxLayout()
        else:
            vl = Qt.QVBoxLayout()
        self.show_labels = show_labels
        self.use_scene_cam = use_scene_cam
        self.debug = debug

        # Set up image-based widgets (i.e. numpy array inputs).
        vl1 = Qt.QHBoxLayout()
        self.image_widge_dict = dict()
        for key in image_keys:
            self.image_widge_dict[key] = Qt.QLabel()
            vl1.addWidget(self.image_widge_dict[key])
        vl.addLayout(vl1)

        # Set up 3D scene widgets (i.e. trimesh scenes).
        self.scene_key_dict = {key: i for i, key in enumerate(scene_keys)}
        self.vps = []
        self.vtkWidgets = []
        vl2 = Qt.QHBoxLayout()
        for _, scene_id in self.scene_key_dict.items():
            vtk_widget = QVTKRenderWindowInteractor(frame)
            vl2.addWidget(vtk_widget)
            vedo_version = importlib.metadata.version('vedo')
            if packaging.version.parse(vedo_version) > packaging.version.parse('2022.3.0'):
                vp = vedo.Plotter(qt_widget=vtk_widget)
            else:
                vp = vedo.Plotter(qtWidget=vtk_widget)  # noqa
            vp.add_callback('LeftButtonPress', partial(self.__on_mouse_start, align_id=scene_id))
            vp.add_callback('LeftButtonRelease', self.__on_mouse_end)
            vp.add_callback('MiddleButtonPress', partial(self.__on_mouse_start, align_id=scene_id))
            vp.add_callback('MiddleButtonRelease', self.__on_mouse_end)
            vp.add_callback('MouseWheelForward', partial(self.__align_event, align_id=scene_id))
            vp.add_callback('MouseWheelBackward', partial(self.__align_event, align_id=scene_id))
            vp.add_callback('KeyPress', self._on_key)
            vp.show()
            self.vtkWidgets.append(vtk_widget)
            self.vps.append(vp)
        vl.addLayout(vl2)

        vl3 = Qt.QVBoxLayout()
        self.figure_key_dict = {key: i for i, key in enumerate(figure_keys)}
        self.figureCanvas = []
        for _ in figure_keys:
            figure_canvas = FigureCanvas(Figure())
            self.figureCanvas.append(figure_canvas)
            #vl3.addWidget(NavigationToolbar(figure_canvas, self))
            vl3.addWidget(figure_canvas)
        vl.addLayout(vl3)

        # Set up mouse timer for synchronizing mouse interactions across multiple (3D) widgets.
        # Therefore, the interaction is detected, tracked and transferred to all other 3D widgets.
        self._mouse_timer = Qt.QTimer(self)
        self._mouse_timer.timeout.connect(self.__align_event)  # noqa
        self._current_align_id = None

        frame.setLayout(vl)
        self.setCentralWidget(frame)
        # If image and scene keys are displayed, somehow the scene widgets start with zero width.
        # Therefore, resize the window before startup by multiplying the current size with the
        # number of scene widgets.
        if len(image_keys) > 0 and len(scene_keys) > 0:
            w = self.size().width() + self.size().width() * len(scene_keys)
            h = self.size().height()
            self.resize(w, h)
        self.show()

        # Set up video directory for storing screenshots.
        self.video_directory = None
        self.__video_index = 0
        if video_directory is not None:
            self.video_directory = pathlib.Path(video_directory)
            if self.video_directory.exists():
                print("\033[93m" + "Video directory already exists, overwriting it ..." + "\033[0m")
            self.video_directory.mkdir(exist_ok=True, parents=True)

    def render_(self, scene_dict: SceneDictType, resetcam: bool = False):
        start_time = time.perf_counter()

        for key, element in scene_dict.items():
            if isinstance(element, trimesh.Scene):
                meshes_vedo, camera_dict = trimesh_scene_2_vedo(
                    scene=element,
                    label=key if self.show_labels else None,
                    use_scene_cam=self.use_scene_cam
                )
                self.render_scene_at_key_(key, meshes_vedo, resetcam=resetcam, camera_dict=camera_dict)

            elif isinstance(element, np.ndarray) and key in self.image_widge_dict:
                if len(element.shape) == 3:
                    if element.shape[0] == 3:
                        element = np.transpose(element, (1, 2, 0))  # (C, H, W) -> (H, W, C)
                else:
                    element = np.stack([element] * 3, axis=-1)
                height, width, _ = element.shape
                qt_img = Qt.QImage(element.data.tobytes(), width, height, 3 * width, Qt.QImage.Format_RGB888)
                self.image_widge_dict[key].setPixmap(Qt.QPixmap.fromImage(qt_img))

            elif isinstance(element, str):
                print(f"{key}: {element}")

            elif isinstance(element, Axes) and key in self.figure_key_dict:
                at = self.figure_key_dict[key]
                # have to use this width/height to adjust the size of the axis somehow
                width, height = self.figureCanvas[at].get_width_height()
                # clear figure
                self.figureCanvas[at].figure.clf()
                # set new references of axes element
                element.figure = self.figureCanvas[at].figure
                self.figureCanvas[at].figure.add_axes(element)
                # draw figure to update!
                self.figureCanvas[at].draw()

            elif isinstance(element, vedo.Volume):
                self.render_scene_at_key_(key, [element], resetcam=resetcam)

            else:
                print("\033[93m" + f"Invalid element in scene dict of type {type(element)}, skipping ..." + "\033[0m")
                continue

        if self.debug:
            dt = int((time.perf_counter() - start_time) * 1000)  # secs -> milliseconds
            print("\033[36m" + f"Rendering time: {dt} ms" + "\033[0m")

        if self.video_directory is not None:
            self.screenshot(self.video_directory, prefix=f"{self.__video_index:05d}")
            self.__video_index += 1

    def render_scene_at_key_(
        self,
        key: str,
        meshes: List[Union[vedo.Mesh, vedo.Volume]],
        resetcam: bool = False,
        camera_dict: Optional[Dict[str, np.ndarray]] = None
    ):
        if key not in self.scene_key_dict:
            print("\033[93m" + f"Key {key} not in scene key dicts, skipping ..." + "\033[0m")
            return

        at = self.scene_key_dict[key]
        self.vps[at].clear()
        self.vps[at].show(meshes, bg="white", resetcam=resetcam, camera=camera_dict)

    def __align_event(self, event=None, align_id: int = None):  # noqa
        if self.num_scenes < 2:
            return
        if align_id is None:
            align_id = self._current_align_id

        camera = self.vps[align_id].camera
        for k in range(self.num_scenes):
            self.vps[k].renderer.SetActiveCamera(camera)
            self.vps[k].render()

    def keyPressEvent(self, event):
        """In case the Qt key callback is called, convert the Qt event to a string key and call the _on_key() method.
        However, this case will only occur if the window contains no vedo panel. Thus, duplicated calling of
        the key callback is not an issue.
        """
        event_dict = {"keyPressed": chr(event.key()).lower()}
        self._on_key(event_dict)
        event.accept()

    def _on_key(self, event_dict) -> None:
        key_pressed = event_dict["keyPressed"]
        if key_pressed == "z":
            if self.num_scenes > 0:
                self.vps[0].render(resetcam=True)
                self.__align_event(align_id=0)
        elif key_pressed == "k":
            screenshot_dir = pathlib.Path("screenshots")
            prefix = datetime.datetime.now().strftime("%Y_%m_%d_%H_%M_%S_%f")
            self.screenshot(screenshot_dir, prefix=prefix)
        elif key_pressed == "h":
            self.print_usage()
        elif key_pressed == "q":
            self.close()

    def __on_mouse_start(self, _, align_id: int):
        self._current_align_id = align_id
        self._mouse_timer.start(10)  # in milliseconds

    def __on_mouse_end(self, _):
        if self._mouse_timer.isActive():
            self._mouse_timer.stop()
        self._current_align_id = None

    @staticmethod
    def print_usage():
        print("Usage:",
              "\n\tq: quit",
              "\n\tz: reset view",
              "\n\tk: take screenshot")

    def screenshot(self, directory: pathlib.Path, prefix: str):
        image_paths = []

        # Store the images from the widget's pixmaps, store them and add them to
        # the list of screenshot files.
        for key, widget in self.image_widge_dict.items():
            path = directory / key / f"{prefix}.png"
            path.parent.mkdir(parents=True, exist_ok=True)
            widget.pixmap().toImage().save(path.as_posix())
            image_paths.append(path)

        # Make screenshots of all trimesh scenes, store them and add them to
        # the list of screenshot files.
        for key, widget in self.scene_key_dict.items():
            path = directory / key / f"{prefix}.png"
            path.parent.mkdir(parents=True, exist_ok=True)
            at = self.scene_key_dict[key]
            self.vps[at].screenshot(path)
            image_paths.append(path)

        # Make to image grid of all screenshot files and store it.
        image_grid_ = image_grid.image_grid(image_paths)[..., :3]  # RGBA -> RGB
        Image.fromarray(image_grid_).save(directory / f"{prefix}__.png")

    def on_close(self):
        for vtk_widget in self.vtkWidgets:
            vtk_widget.close()
        for key, widget in self.image_widge_dict.items():
            widget.close()

    @property
    def num_images(self) -> int:
        return len(self.image_widge_dict)

    @property
    def num_scenes(self) -> int:
        return len(self.vps)



import numpy as np
import trimesh
import trimesh.visual
import vedo
import vtk

from typing import Dict, List, Optional, Tuple
from vtk.util.numpy_support import numpy_to_vtk


def trimesh_scene_2_vedo(scene: trimesh.Scene, label: Optional[str] = None, use_scene_cam: bool = False
                         ) -> Tuple[List[vedo.Mesh], Optional[Dict[str, np.ndarray]]]:
    meshes_vedo = []
    for m in scene.geometry.values():
        if isinstance(m, trimesh.Trimesh):
            m_vedo = vedo.trimesh2vedo(m)
            if m.visual.kind == "vertex":
                face_colors = trimesh.visual.color.vertex_to_face_color(m.visual.vertex_colors, m.faces)
                m_vedo.cellIndividualColors(face_colors)
            meshes_vedo.append(m_vedo)

        elif isinstance(m, trimesh.PointCloud):
            vertices = m.vertices
            vertex_colors = m.visual.vertex_colors
            n, _ = vertices.shape

            # vedo.Points uses vtk.vtkPolyData() as backend data storage and thus converts
            # the input to this type. However, the vedo conversion is quite inefficient, therefore
            # we convert the trimesh.PointCloud before passing it to vedo.
            # partially from https://github.com/pyvista/utilities/helpers.py
            vtkpts = vtk.vtkPoints()
            vtk_arr = numpy_to_vtk(vertices, deep=True)
            vtkpts.SetData(vtk_arr)
            pd = vtk.vtkPolyData()
            pd.SetPoints(vtkpts)

            # For some reason, vedo requires each vtk.vtkPolyData object to set an internal
            # cell array, as it uses the vertices, not the points attribute. As a point cloud
            # is unconnected, the offset and connectivity is just the index of the respective point.
            carr = vtk.vtkCellArray()
            carr.SetData(vedo.numpy2vtk(np.arange(n + 1), dtype="int"),  # offset
                         vedo.numpy2vtk(np.arange(n), dtype="int"))  # connectivity
            pd.SetVerts(carr)

            # Set vertex color RGB/RGBA values as active scalar property of the vtk.vtkPolyData.
            if vertex_colors.shape[1] == 3:
                ucols = numpy_to_vtk(vertex_colors)
                ucols.SetName("Points_RGB")
                pd.GetPointData().AddArray(ucols)
                pd.GetPointData().SetActiveScalars("Points_RGB")
            elif vertex_colors.shape[1] == 4:
                ucols = numpy_to_vtk(vertex_colors)
                ucols.SetName("Points_RGBA")
                pd.GetPointData().AddArray(ucols)
                pd.GetPointData().SetActiveScalars("Points_RGBA")
            else:
                print("\033[93m" + f"Invalid point cloud colors, skipping ..." + "\033[0m")

            m_vedo = vedo.Points(pd)
            meshes_vedo.append(m_vedo)

        elif isinstance(m, trimesh.path.Path3D):
            # The trimesh path consists of entities and vertices. The vertices are the 3D points,
            # that are connected as described in the entities.
            if not all([isinstance(me, trimesh.path.entities.Line) for me in m.entities]):
                raise ValueError("Currently only trimesh.path.entities.Line entities are supported")
            if not all([len(me.points) == 2 for me in m.entities]):
                raise ValueError("Invalid line entities, should have point lists [start, end]")

            # Add each line segment individually as a vedo line to support multicolored lines
            # and different alpha values along the line.
            for ke, line_entity in enumerate(m.entities):
                i, j = line_entity.points
                c = m.colors[ke, :3]
                alpha = m.colors[ke, -1] / 255  # [0, 255] -> [0, 1]
                m_vedo = vedo.Lines(m.vertices[None, i], m.vertices[None, j], lw=2, c=c, alpha=alpha)
                meshes_vedo.append(m_vedo)

        else:
            meshes_vedo.append(m)

    if label is not None:
        annotation = vedo.CornerAnnotation()
        annotation.text(label)
        meshes_vedo.append(annotation)

    camera_dict = None
    if use_scene_cam:
        focal_distance = 1.0
        T_W_C = scene.camera_transform
        cam_0 = T_W_C[:3, 3]
        cam_1 = cam_0 + T_W_C[:3, :3] @ np.array([0, 0, focal_distance])  # along z of camera
        view_up = - T_W_C[:3, 1]  # camera convention -> y points down
        camera_dict = dict(pos=cam_0, focal_point=cam_1, viewup=view_up)

    return meshes_vedo, camera_dict



import pathlib
import pickle as pkl

from functools import partial
from PyQt5 import Qt
from PyQt5.QtCore import QObject, QThread, pyqtSignal
from typing import List, Optional, Union

from .window_base import MainWindow, SceneDictType


__all__ = ['display_real_time', 'RealTimeNode']


class RealTimeNode(QObject):
    """Template class for real time rendering.

    Examples:
        class DemoNode(RealTimeNode):

            def callback(self):
                scene = trimesh.Scene()
                box_mesh = trimesh.creation.box(bounds=[(-1, -1, 0), (1, 1, 1)])
                scene.add_geometry(box_mesh)
                self.render({"scene": scene})

                if condition:
                    self.close()
    """
    is_finished = False
    running = pyqtSignal()
    scene_dict_emitter = pyqtSignal(dict)

    def callback(self):
        raise NotImplementedError

    def run(self):
        while not self.is_finished:
            if not self.running:
                continue
            self.callback()

    def render(self, scene_dict: SceneDictType):
        self.scene_dict_emitter.emit(scene_dict)  # noqa

    def stop(self):
        self.is_finished = True


class CachingSignals(QObject):
    finished = pyqtSignal()


class CachingNode(Qt.QRunnable):

    def __init__(self, file_path: pathlib.Path, scene_dict: SceneDictType):
        super(CachingNode, self).__init__()
        self.file_path = file_path
        self.scene_dict = scene_dict
        self.signals = CachingSignals()

    @Qt.pyqtSlot()
    def run(self):
        with open(self.file_path, 'wb+') as f:
            pkl.dump(self.scene_dict, file=f)
        self.signals.finished.emit()


class MainWindowRealTime(MainWindow):

    def __init__(
        self,
        worker: QObject,
        image_keys: List[str],
        scene_keys: List[str],
        figure_keys: List[str],
        video_directory: Optional[Union[pathlib.Path, str]] = None,
        cache_directory: Optional[Union[pathlib.Path, str]] = None,
        horizontal: bool = True,
        show_labels: bool = False,
        use_scene_cam: bool = False,
        debug: bool = False,
        parent: Qt.QWidget = None
    ):
        super(MainWindowRealTime, self).__init__(
            image_keys=image_keys,
            scene_keys=scene_keys,
            figure_keys=figure_keys,
            video_directory=video_directory,
            horizontal=horizontal,
            show_labels=show_labels,
            use_scene_cam=use_scene_cam,
            debug=debug,
            parent=parent
        )

        self.cache_directory = None
        self.__cache_index = 0
        self.cache_thread_pool = None
        if cache_directory is not None:
            self.cache_directory = pathlib.Path(cache_directory)
            if self.cache_directory.exists():
                print("\033[93m" + "Cache directory already exists, overwriting it ..." + "\033[0m")
            self.cache_directory.mkdir(exist_ok=True, parents=True)
            self.cache_thread_pool = Qt.QThreadPool()
            if self.debug:
                caching_max_threads = self.cache_thread_pool.maxThreadCount()
                print("\033[36m" + f"Launched caching thread pool with max. {caching_max_threads} threads" + "\033[0m")

        # Set up multi-processing pipeline.
        self.thread = QThread(parent=self)
        self.worker = worker
        self.worker.moveToThread(self.thread)
        self.thread.started.connect(self.worker.run)
        self.worker.scene_dict_emitter.connect(self.render_)
        self.thread.start()

        self.show()

    def render_(self, scene_dict: SceneDictType, resetcam: bool = False):
        super(MainWindowRealTime, self).render_(scene_dict, resetcam=resetcam)

        # If a cache directory is defined, write the scene dict to a pickle file, one per scene_dict.
        if self.cache_directory is not None:
            cache_file_name = self.cache_directory / f"{self.__cache_index:05d}.pkl"
            cache_node = CachingNode(cache_file_name, scene_dict)
            cache_node.signals.finished.connect(partial(self.__caching_complete, cache_file_name))
            self.cache_thread_pool.start(cache_node)
            self.__cache_index += 1

    def __caching_complete(self, cache_file_name: pathlib.Path):
        if self.debug:
            print("\033[36m" + f"Caching complete: {cache_file_name.as_posix()}" + "\033[0m")

    def _on_key(self, event_dict) -> None:
        super(MainWindowRealTime, self)._on_key(event_dict)
        key_pressed = event_dict["keyPressed"]
        if key_pressed == "s":
            self.worker.running = not self.worker.running

    @staticmethod
    def print_usage():
        MainWindow.print_usage()
        print("\ts: play / pause")

    def on_close(self):
        super(MainWindowRealTime, self).on_close()
        self.worker.stop()
        self.thread.quit()
        self.thread.wait()
        if self.cache_thread_pool is not None:
            self.cache_thread_pool.waitForDone()


def display_real_time(
    worker: RealTimeNode,
    image_keys: Optional[List[str]] = None,
    scene_keys: Optional[List[str]] = None,
    figure_keys: List[str] = None,
    horizontal: bool = True,
    video_directory: Optional[Union[pathlib.Path, str]] = None,
    cache_directory: Optional[Union[pathlib.Path, str]] = None,
    show_labels: bool = False,
    use_scene_cam: bool = False,
    debug: bool = False
):
    """Display scenes from a real-time application, i.e. render in-coming scenes directly.

    The worker is a process that generates each new scene in a separate thread and passes it to the renderer.
    For a smooth rendering process and for simplification, the image and scene keys have to be known in advances
    and cannot be changed afterwards.

    Args:
        worker: generator of new scenes.
        image_keys: names of images in rendered dictionary.
        scene_keys: names of 3D scenes in rendered dictionary.
        figure_keys: names of matplotlib figures in rendered dictionary.
        horizontal: window orientation, horizontal or vertical stacking.
        video_directory: directory for storing screenshots.
        cache_directory: directory for caching the input scene dicts for re-rendering.
        show_labels: display the scene dict keys.
        use_scene_cam: use camera transform from trimesh scenes.
        debug: printing debug information.
    """
    if image_keys is None and scene_keys is None:
        raise ValueError("Neither image nor scene keys, provide at least one key!")
    if not hasattr(worker, 'run'):
        raise ValueError("Worker must have method run()")
    if image_keys is None:
        image_keys = []
    if scene_keys is None:
        scene_keys = []
    if figure_keys is None:
        figure_keys = []

    app = Qt.QApplication([])
    window = MainWindowRealTime(
        worker=worker,
        image_keys=image_keys,
        scene_keys=scene_keys,
        figure_keys=figure_keys,
        video_directory=video_directory,
        cache_directory=cache_directory,
        horizontal=horizontal,
        show_labels=show_labels,
        use_scene_cam=use_scene_cam,
        debug=debug
    )
    app.aboutToQuit.connect(window.on_close)
    app.exec_()



import numpy as np

from typing import Tuple, Union


def create_round_flight(
    timestamps: np.ndarray,
    focal_point: Union[Tuple[float, float, float], np.ndarray] = (0, 0, 0),
    cam_height: float = 2.0,
    radius: float = 5.0
):
    """Create camera trajectory circling around an object in the focal point, in the xy-plane on z = cam_height
    and with the given radius.

    Args:
        timestamps: trajectory timestamps in radians (N,) -> 2*pi = full circle.
        focal_point: trajectory focal point, i.e. each camera points at this point.
        cam_height: camera z coordinate.
        radius: circle radius.
    Returns:
        T_W_C: world to camera transforms (N, 4, 4).
    """
    if len(timestamps.shape) != 1:
        raise ValueError(f"Expected flat timestamp vector, got {timestamps.shape}")

    num_time_stamps = len(timestamps)
    cam_t = np.stack([focal_point[0] + radius * np.cos(timestamps),
                      focal_point[1] + radius * np.sin(timestamps),
                      focal_point[2] + np.ones_like(timestamps) * cam_height], axis=1)

    T_W_C = np.zeros((num_time_stamps, 4, 4))
    focal_point_ = np.array(focal_point)
    for k in range(num_time_stamps):
        cam_z = focal_point_ - cam_t[k]
        cam_z = cam_z / np.linalg.norm(cam_z)
        cam_x = np.cross(np.array([0, 0, 1]), cam_t[k])
        cam_x = cam_x / np.linalg.norm(cam_x)
        cam_y = np.cross(cam_z, cam_x)

        T_W_C[k] = np.eye(4)
        T_W_C[k, :3, 3] = cam_t[k]
        T_W_C[k, :3, :3] = np.stack([cam_x, cam_y, cam_z]).T

    return T_W_C



import argparse
import pathlib
import pickle as pkl

from balanna.window_dataset import display_dataset


def _parse_args():
    parser = argparse.ArgumentParser()
    parser.add_argument("directory", type=pathlib.Path, help="cached data directory")
    parser.add_argument("--fps", type=int, help="displaying fps", default=10)
    parser.add_argument("--print-frame-index", action="store_true")
    return parser.parse_args()


def main(args):
    if not args.directory.exists():
        raise NotADirectoryError(f"Input directory {args.directory} not found")

    files = sorted(list(args.directory.glob("*.pkl")))
    if len(files) == 0:
        raise FileNotFoundError(f"No .pkl files found, invalid or empty cache directory")

    for k, file in enumerate(files):
        with open(file, 'rb') as f:
            scene_dict = pkl.load(f)
        if args.print_frame_index and "frame_index" not in scene_dict.keys():
            scene_dict["frame_index"] = str(k)
        yield scene_dict


if __name__ == '__main__':
    args_ = _parse_args()
    display_dataset(main(args_), fps=args_.fps)



import numpy as np
import trimesh
import vedo

from typing import List

from .utils.vedo_utils import trimesh_scene_2_vedo
from .window_base import SceneDictType


__all__ = ["render_dataset"]


def render_dataset(
    scenes: List[SceneDictType],
    show_labels: bool = False,
    use_scene_cam: bool = False,
    debug: bool = False
):
    """Render scenes stored in scene iterator to a numpy array (num_frames, H, W, 3).

    The scene iterator yields a dictionary that describes the elements of the scene, one dictionary per frame.
    See SceneDictType for currently supported object types.

    Args:
        scenes: sorted list of scene dictionaries.
        show_labels: display the scene dict keys.
        use_scene_cam: use camera transform from trimesh scenes.
        debug: printing debug information.
    """
    if len(scenes) == 0:
        return np.zeros((0, 640, 480, 3), dtype=np.uint8)

    num_scenes = [len(scene_dict) for scene_dict in scenes]
    assert num_scenes.count(num_scenes[0]) == len(num_scenes)  # all scenes have same length
    num_scenes = num_scenes[0]  # length of scene dict

    plotter = vedo.Plotter((num_scenes, 1), offscreen=True, interactive=False)
    images = []
    for k, scene_dict in enumerate(scenes):
        plotter.clear()
        for i, (key, scene) in enumerate(scene_dict.items()):
            if not isinstance(scene, trimesh.Scene):
                print("\033[93m" + f"Only trimesh.Scene type is supported for rendering, skipping {key}..." + "\033[0m")
            meshes_vedo, camera_dict = trimesh_scene_2_vedo(
                scene=scene,
                label=key if show_labels else None,
                use_scene_cam=use_scene_cam
            )
            plotter.at(i).show(meshes_vedo, camera=camera_dict)
        image_k = plotter.screenshot(asarray=True)
        images.append(image_k)
        if debug:
            print("\033[36m" + f"Finished frame {k} with shape {image_k.shape}" + "\033[0m")

    plotter.close()
    return np.stack(images)



import os
import numpy as np
import trimesh
import trimesh.creation

from typing import List, Optional, Tuple, Union


def show_mesh(
    vertices: np.ndarray,
    faces: np.ndarray,
    transform: np.ndarray = np.eye(4),
    name: Optional[str] = None,
    scene: Optional[trimesh.Scene] = None,
    face_color: Optional[np.ndarray] = None,
    vertex_color: Optional[np.ndarray] = None
) -> trimesh.Scene:
    """Create a mesh from vertices and faces and add it to the scene.

    Args:
        vertices: mesh vertices (N, 3).
        faces: mesh faces.
        transform: transform mesh before adding it to scene (4, 4).
        name: scene node name of mesh.
        scene: scene to add mesh to, if None a new scene will be created.
        face_color: face colors (F, 3).
        vertex_color: vertex colors (N, 3/4).
    """
    if vertex_color is not None and vertex_color.shape[:-1] != vertices.shape[:-1]:
        raise ValueError(f"Invalid vertex colors, must be {vertices.shape}, got {vertex_color.shape}")
    if vertex_color is not None and face_color is not None:
        raise ValueError(f"Cannot set both vertex and face color")

    if scene is None:
        scene = trimesh.Scene()

    mesh = trimesh.Trimesh(vertices, faces)
    if face_color is not None:
        mesh.visual.face_colors[:, :3] = face_color
    elif vertex_color is not None:
        mesh.visual.vertex_colors = vertex_color
    else:
        mesh.visual.face_colors[:, :3] = [224, 120, 120]

    mesh = mesh.apply_transform(transform)
    scene.add_geometry(mesh, node_name=name)
    return scene


def show_grid(
    xy_min: float = -10.0,
    xy_max: float = 10.0,
    z: float = 0.0,
    resolution: float = 1.0,
    alpha: float = 255,
    dark_color: Tuple[float, float, float] = (120, 120, 120),
    light_color: Tuple[float, float, float] = (255, 255, 255),
    transform: np.ndarray = np.eye(4),
    scene: trimesh.Scene = None
) -> trimesh.Scene:
    """Create a 2D squared grid in the xy-plane, color it in a chessboard pattern and transform it accordingly.

    Args:
        xy_min: minimal coordinate in xy direction, i.e. (xy_min, xy_min).
        xy_max: maximal coordinate in xy direction, i.e. (xy_max, xy_max).
        z: plane height at creation.
        resolution: cell size.
        alpha: color alpha value for all cells.
        dark_color: chessboard dark color.
        light_color: chessboard light color.
        transform: mesh transform after creation (in relation to center point (0, 0, z)).
        scene: scene to add mesh to, if None a new scene will be created.
    """
    if scene is None:
        scene = trimesh.Scene()

    num_points_per_row = int((xy_max - xy_min) // resolution) + 1
    w = num_points_per_row - 1
    num_faces = w ** 2

    # Compute vertices as 2D mesh-grid in the xy-plane. Set all z components to the given z value.
    vertices_x, vertices_y = np.meshgrid(np.linspace(xy_min, xy_max, num_points_per_row),
                                         np.linspace(xy_min, xy_max, num_points_per_row))
    vertices_x = vertices_x.flatten()
    vertices_y = vertices_y.flatten()
    vertices = np.stack([vertices_x, vertices_y, np.ones_like(vertices_x) * z], axis=-1)

    # Compute grid faces and face colors to create chessboard pattern.
    faces = []
    face_colors = np.zeros((num_faces, 4), dtype=np.uint8)
    face_colors[:, -1] = alpha  # alpha value
    sym = [light_color, dark_color]
    face_idx = 0
    for i in range(w):
        for j in range(w):
            # add face in counter-clockwise vertex order.
            face_ij = (i * num_points_per_row + j,
                       i * num_points_per_row + j + 1,
                       (i + 1) * num_points_per_row + j + 1,
                       (i + 1) * num_points_per_row + j)
            faces.append(face_ij)

            # if even number of elements in row: at beginning of row, swap colors.
            # if odd: order is automatically swapped after finishing one row.
            if face_idx % w == 0 and w % 2 == 0:
                sym.reverse()
            idx = 0 if face_idx % 2 == 0 else 1  # alternate index for chessboard pattern
            face_colors[face_idx, :3] = sym[idx]

            # Increment face index.
            face_idx += 1
    face_colors = np.concatenate([face_colors, face_colors])

    grid = trimesh.Trimesh(vertices, faces, face_colors=face_colors)
    scene.add_geometry(grid, transform=transform)
    return scene


def show_point_cloud(
    vertices: np.ndarray,
    transform: np.ndarray = np.eye(4),
    colors: Optional[np.ndarray] = None,
    name: Optional[str] = None,
    scene: Optional[trimesh.Scene] = None
) -> trimesh.Scene:
    """Add vertices as trimesh point cloud to the scene.

    Trimesh expects a flat point cloud of shape (N, 3). Therefore, if the point cloud is not flat, it will be
    flattened before passing it to trimesh.

    Args:
        vertices: point cloud to visualize (..., 3).
        transform: transform point cloud before adding it to scene (4, 4).
        colors: RGB color for each point (..., 3) as uint8 color 0...255..
        name: scene node name of point cloud mesh.
        scene: scene to add point cloud mesh to, if None a new scene will be created.
    """
    if scene is None:
        scene = trimesh.Scene()
    if transform.shape != (4, 4):
        raise ValueError(f"Expected single transformation matrix, got {transform.shape}")

    vertices_flat = vertices.reshape(-1, 3)
    if colors is None:
        colors = np.zeros((len(vertices_flat), 3))
        colors[:, 0] = 255  # default = red
    pc = trimesh.PointCloud(vertices_flat, colors=colors)
    scene.add_geometry(pc, transform=transform, node_name=name)
    return scene


def show_trajectory(
    trajectory: np.ndarray,
    colors: Tuple[float, float, float] = (255, 0, 0),
    alpha: int = 255,
    fade_out: bool = False,
    scene: Optional[trimesh.Scene] = None
) -> trimesh.Scene:
    """Add a colored trajectory to the scene

    Args:
        trajectory: path to be drawn (N, 3).
        colors: path color (uniformly colored line) as RGB tuple.
        alpha: path alpha value. If fade_out, max alpha value.
        fade_out: fade out line by linearly decreasing the alpha value along the path length.
        scene: scene to add path to, if None a new scene will be created.
    """
    if len(trajectory.shape) != 2 or trajectory.shape[-1] != 3:
        raise ValueError(f"Invalid size of trajectory, expected (N, 3), got {trajectory.shape}")
    if scene is None:
        scene = trimesh.Scene()

    # If the trajectory just contains a single point, then it cannot be drawn. Just return the scene.
    if trajectory.shape[0] < 2:
        print("\033[93m" + "Trajectory must contain at least two points, skipping ..." + "\033[0m")
        return scene

    # In trimesh a path consists of line segments, each connecting one vertex with the next one. The
    # way of connection is determined from the entities (similar to faces in 3D meshes).
    # Each line segment could have a different color, but vedo just supports uniformly colored lines,
    # which would be supported, but over-complicate things here.
    # process = False, otherwise vertices are weirdly reconnected
    num_points = trajectory.shape[0]
    entities = [trimesh.path.entities.Line([j, j + 1]) for j in range(num_points - 1)]

    if fade_out:
        entity_colors = [(*colors, alpha * ((k + 2) / num_points)) for k in range(num_points - 1)]
    else:
        entity_colors = [(*colors, alpha) for _ in range(num_points - 1)]
    path = trimesh.path.Path3D(entities=entities, vertices=trajectory, colors=entity_colors, process=False)

    scene.add_geometry(path)
    return scene


def show_camera(
    transform: np.ndarray,
    name: Optional[str] = None,
    scene: Optional[trimesh.Scene] = None
) -> trimesh.Scene:
    """Load camera mesh and transform it to given pose.

    Args:
        transform: camera pose as transformation matrix (4, 4).
        name: scene node name of the camera mesh.
        scene: scene to add camera to, if None a new scene will be created.
    """
    if scene is None:
        scene = trimesh.Scene()
    if transform.shape != (4, 4):
        raise ValueError(f"Expected single transformation matrix, got {transform.shape}")
    script_path = os.path.basename(os.path.dirname(__file__))
    cam_mesh_path = os.path.join(script_path, "..", "meshes", "cam_z.obj")
    cam_mesh = trimesh.load_mesh(cam_mesh_path)  # already in camera orientation (z forward, x right, y down)
    cam_mesh = cam_mesh.apply_transform(transform)
    scene.add_geometry(cam_mesh, node_name=name)
    return scene


def show_axis(
    transform: np.ndarray,
    name: Optional[str] = None,
    size: float = 0.06,
    scene: Optional[trimesh.Scene] = None
) -> trimesh.Scene:
    """Add coordinate axis as trimesh mesh to the scene.

    Args:
        transform: axis pose as transformation matrix (4, 4).
        name: scene node name of axis mesh.
        size: axis origin size and radius.
        scene: scene to add axis to, if None a new scene will be created.
    """
    if scene is None:
        scene = trimesh.Scene()
    if transform.shape != (4, 4):
        raise ValueError(f"Expected single transformation matrix, got {transform.shape}")
    axis_mesh = trimesh.creation.axis(size, transform=transform, axis_radius=size/3)
    scene.add_geometry(axis_mesh, node_name=name)
    return scene


def show_axes(
    transforms: Union[List[np.ndarray], np.ndarray],
    sizes: Optional[List[float]] = None,
    scene: Optional[trimesh.Scene] = None
):
    """Add coordinate axes for multiple frames to the scene.

    Args:
        transforms: axes poses as transformation matrices (N, 4, 4 or list of (4, 4)).
        sizes: axes sizes.
        scene: scene to add axis to, if None a new scene will be created.
    """
    if sizes is None:
        sizes = [0.04] * len(transforms)
    for tf, size in zip(transforms, sizes):
        scene = show_axis(tf, size=size, scene=scene)
    return scene



from setuptools import setup, find_packages


REQUIRED = [
    'numpy>=1.21',
    'trimesh[easy]>=3.16.0',
    'vedo',
    'PyQt5',
    'image_grid'
    # 'opencv-python-headless'
]

setup(
    name='balanna',
    version='1.2',
    description='Visualization Tools for 2D & 3D stuff working out-of-the-box',
    author='Simon Schaefer & Dorian Henning',
    author_email='simon.k.schaefer@gmail.com',
    packages=find_packages(),
    install_requires=REQUIRED,
    python_requires='>=3.8'
)



import numpy as np
import trimesh


def quaternion_to_trimesh_format(quaternion: np.ndarray, q_order: str = "XYZW"):
    if q_order == "XYZW":
        return quaternion[..., [3, 0, 1, 2]]
    elif q_order == "WXYZ":
        return quaternion
    raise ValueError(f"Unknown quaternion config {q_order}")


def pose_to_trimesh_tf(position: np.ndarray, quaternion: np.ndarray, q_order: str = "XYZW"):
    q = quaternion_to_trimesh_format(quaternion, q_order=q_order)
    T = trimesh.transformations.quaternion_matrix(q)
    T[..., :3, 3] = position
    return T



import numpy as np
import pathlib

from matplotlib.axes import Axes
from PyQt5 import Qt
from typing import Iterable, Optional, Union

from .window_base import MainWindow, SceneDictType
from .utils.types import contains_scene


__all__ = ['display_scenes', 'display_generated']


class MainWindowGenerator(MainWindow):

    def __init__(
        self,
        scene_iterator: Iterable[SceneDictType],
        fps: float,
        video_directory: Optional[Union[pathlib.Path, str]] = None,
        horizontal: bool = True,
        loop: bool = True,
        show_labels: bool = False,
        use_scene_cam: bool = False,
        debug: bool = False,
        parent: Qt.QWidget = None
    ):
        self.scene_iterator = scene_iterator
        self.fps = fps
        scene_dict = self.get_next_scene_dict()
        if scene_dict is None:
            return

        image_keys = [key for key, value in scene_dict.items() if isinstance(value, np.ndarray)]
        scene_keys = [key for key, value in scene_dict.items() if contains_scene(value)]
        figure_keys = [key for key, value in scene_dict.items() if isinstance(value, Axes)]
        super(MainWindowGenerator, self).__init__(
            image_keys=image_keys,
            scene_keys=scene_keys,
            figure_keys=figure_keys,
            video_directory=video_directory,
            horizontal=horizontal,
            show_labels=show_labels,
            use_scene_cam=use_scene_cam,
            debug=debug,
            parent=parent
        )

        self.render_(scene_dict, resetcam=True)

        # Setup looping timer and connect it to render a new scene at every timer callback.
        # If loop is true, start looping right away.
        self.timer = Qt.QTimer(self)
        self.timer.timeout.connect(self.render_next_scene)  # noqa
        if loop:
            self.toggle_looping()

    def render_next_scene(self, resetcam: bool = False) -> None:
        scene_dict = self.get_next_scene_dict()
        if scene_dict is not None:
            self.render_(scene_dict, resetcam=resetcam)
        elif self.timer.isActive():
            self.timer.stop()

    def _on_key(self, event_dict) -> None:
        super(MainWindowGenerator, self)._on_key(event_dict)
        key_pressed = event_dict["keyPressed"]
        if key_pressed == "s":
            self.toggle_looping()
        elif key_pressed == "n":
            self.render_next_scene()

    @staticmethod
    def print_usage():
        MainWindow.print_usage()
        print("\ts: play / pause",
              "\n\tn: next")

    def get_next_scene_dict(self) -> Optional[SceneDictType]:
        try:
            return next(self.scene_iterator)  # noqa
        except StopIteration:
            return None

    def toggle_looping(self):
        if self.timer.isActive():
            self.timer.stop()
        else:
            self.timer.start(int(1 / self.fps * 1000))  # in milliseconds


def display_generated(
    scene_iterator: Iterable[SceneDictType],
    horizontal: bool = True,
    loop: bool = False,
    fps: float = 30.0,
    video_directory: Optional[Union[pathlib.Path, str]] = None,
    show_labels: bool = False,
    use_scene_cam: bool = False,
    debug: bool = False
):
    """Display scenes stored in scene iterator as PyQt app.

    The scene iterator yields a dictionary that describes the elements of the scene, one dictionary per frame.
    See SceneDictType for currently supported object types.

    Args:
        scene_iterator: iterator function to get the scene dictionaries.
        horizontal: window orientation, horizontal or vertical stacking.
        loop: start looping from beginning.
        fps: frames (i.e. scenes) per second for looping.
        video_directory: directory for storing screenshots.
        show_labels: display the scene dict keys.
        use_scene_cam: use camera transform from trimesh scenes.
        debug: printing debug information.
    """
    app = Qt.QApplication([])
    window = MainWindowGenerator(
        scene_iterator=scene_iterator,
        fps=fps,
        video_directory=video_directory,
        horizontal=horizontal,
        loop=loop,
        show_labels=show_labels,
        debug=debug,
        use_scene_cam=use_scene_cam
    )
    app.aboutToQuit.connect(window.on_close)
    app.exec_()


def display_scenes(*args, **kwargs):
    print("\033[93m" + f"display_scenes() is out-dated and will be removed in the next version, "
                       f"use display_generated() instead" + "\033[0m")
    return display_generated(*args, **kwargs)
